(function ($) {
	$.widget("lws.lac_model", {
		/** LWS Autocomplete Control Model
		 * Model in charge of datasource management (direct/ajax) and research / Autocomplete inside the source
		 *
		 * The direct data source can be set in multiple ways :
		 * @param <option> -> <option> in select element (Lowest Priority)
		 * @param origin.options.source -> Array of Objects (value, label, parent, html) (Medium Priority)
		 * @param data-source -> base64 encoded array of objects in DOM Element (Highest Priority)
		 *
		 * The remote data source can be set in multiple ways :
		 * @param origin.options.sourceurl -> direct url to JSON file (Lowest Priority)
		 * @param origin.options.ajax -> action to perform in the wordpress ajax service (Lowest Priority)
		 * or within a DOM Element using data attributes :
		 * @param data-sourceurl -> direct url to JSON file (Highest Prority)
		 * @param data-ajax -> the action of the wordpress ajax service, if none, the <option> list is used as data source.
		 *
		 * Other options
		 * @param origin.options.minsearch or @param data-minsearch -> the minimal search string length before ajax call instead of local options.
		 * @param origin.options.minoption or @param data-minoption -> if local filter result count is less or equal, ajax call (if any) is attempt.
		 * @param origin.options.minlength or @param data-minlength -> minimal input length before autocomplete starts (default 1).
		 *
		 * @param mode -> 'autocomplete' (match from beginning) or 'research' (search within)
		 * @param origin -> Calling widget
		 * @param resChange -> Callback function to send the results to
		 * @param resLoad -> Callback function to call when ajax is loading
		 */
		options: {
			mode: 'autocomplete',
			origin: undefined,
			count: 20,
		},

		_create: function () {
			this.originsList = [];
			this.currentOrigin = {};
			this.dataSource = new Map();
			this.flatSource = new Map();
			this.groupSource = new Map();
			this.showMore = false;
			$.ajaxSettings.cache = false;
			this._setOrigin(this.options.origin);
			this._setSource();
		},

		_bind: function (fn, me, data) {
			return function () {
				return fn.apply(me, data);
			};
		},

		/* Go through all options to set the data sources */
		_setSource: function () {
			// direct data source
			var domObject = $(this.options.origin.element);
			this.dataSource = new Map();
			this.flatSource = new Map();
			this.groupSource = new Map();
			this.ajax = {};
			if (domObject[0].nodeName == "SELECT") {
				for (var i = 0; i < domObject[0].options.length; i++) {
					var dataLine = {};
					dataLine.value = domObject[0].options[i].value;
					dataLine.label = domObject[0].options[i].text;
					if (domObject[0].options[i].closest('optgroup') != null) {
						var groupName = domObject[0].options[i].closest('optgroup').label;
						if (!this.dataSource.has(groupName)) {
							var grp = {
								'label': groupName,
								'group': new Map()
							};
							this.dataSource.set(groupName.toString(), grp);
							this.groupSource.set(groupName.toString(), grp);
						}
						this.dataSource.get(groupName.toString()).group.set(dataLine.value.toString(), dataLine);
						this.flatSource.set(dataLine.value.toString(), dataLine);
					} else {
						this.dataSource.set(dataLine.value.toString(), dataLine);
						this.flatSource.set(dataLine.value.toString(), dataLine);
					};
				}
			};
			if (this.options.origin.options.source != undefined) {
				this._updateSource(this.dataSource, this.options.origin.options.source)
			}
			if (domObject.data('source') != undefined && domObject.data('source') != '') {
				this._updateSource(this.dataSource, lwsBase64.toObj(domObject.data('source')))
			}

			// ajax data source
			if (this.options.origin.options.ajax != undefined) {
				this.showMore = true;
				this.ajax.action = this.options.origin.options.ajax;
			}
			if (domObject.data('ajax') != undefined) {
				this.showMore = true;
				this.ajax.action = domObject.data('ajax');
				this._remoteSource('', 'force');
			}
			if (this.options.origin.options.spec != undefined)
				this.ajax.spec = this.options.origin.options.spec;
			if (domObject.data('sourceurl') != undefined)
				this.ajax.sourceurl = domObject.data('sourceurl');

			// other options
			if (this.options.origin.options.minsearch != undefined)
				this.options.minsearch = this.options.origin.options.minsearch;
			if (domObject.data('minsearch') != undefined)
				this.options.minsearch = domObject.data('minsearch');
			if (this.options.origin.options.minoption != undefined)
				this.options.minoption = this.options.origin.options.minoption;
			if (domObject.data('minoption') != undefined)
				this.options.minoption = domObject.data('minoption');

			if (this.options.origin.options.minlength != undefined)
				this.options.minlength = this.options.origin.options.minlength;
			if (domObject.data('minlength') != undefined)
				this.options.minlength = domObject.data('minlength');
		},

		/* Set/reset Origin parameters for multiple calling widgets on same model */
		_setOrigin: function (origin) {
			var uuid = origin.uuid;
			//New origin
			if (this.originsList[uuid] == '' || this.originsList[uuid] == undefined) {
				this.originsList[uuid] = {};
				this.originsList[uuid].resLoad = origin.resLoad;
				this.originsList[uuid].resChange = origin.resChange;
				this.originsList[uuid].target = origin;
				this.originsList[uuid].searched = '';
				this.originsList[uuid].page = 0;
				this.originsList[uuid].count = this.options.count;
			}
			this.currentOrigin = this.originsList[uuid];
		},

		/* Reinit the DataSource */
		reInit: function (origin) {
			this._setOrigin(origin);
			this.currentOrigin.page = 0;
			this._setSource();
		},

		/* Show More ajax results */
		showMoreResults: function (origin) {
			this._setOrigin(origin);
			this.currentOrigin.page += 1;
			this._remoteSource(this.currentOrigin.searched);
		},

		/* Send back labels corresponding to initial values */
		returnLabels: function (initData, origin, mode) {
			this.resCount = 0;
			if (origin != undefined && origin != '')
				this._setOrigin(origin);
			result = [];

			if (initData != undefined && initData !== false && initData.length > 0 && initData[0] != '' && initData[0] != null) {
				notFound = [];
				result = this._flatSearch(initData, 'value', notFound);
				if (notFound.length > 0) {
					if ((this.ajax.sourceurl != undefined && this.ajax.sourceurl != '') || (this.ajax.action != undefined && this.ajax.action != '')) {
						this._remoteSource(initData, "init", true, true);
					}
					/* While waiting for the remote Source or if not in remote -> create Data */
					for (var i = 0; i < notFound.length; i++) {
						res = {
							'value': notFound[i].toString(),
							'label': notFound[i],
							'html': ''
						};
						result.push(res);
						this._updateSource(this.dataSource, res);
					}
				}
			}
			this._sendBack(result, "init")
		},

		getLabelFromValue: function (value) {
			var notFound = [];
			var result = this._flatSearch([value], 'value', notFound);
			return result.length ? result[0].label : '';
		},

		/* returns true if the searchString is a substring of previous search */
		isSubChain: function (searchString, origin) {
			if (origin != undefined && origin != '') this._setOrigin(origin);
			if (this.currentOrigin.searched.toLowerCase().substring(0, searchString.length) == searchString.toLowerCase() && searchString.length < this.currentOrigin.searched.length) return true;
			return false;
		},

		/** Send back values corresponding to sent labels or assign labels to values
		 * @param readOnly (optional: default=false) if true, do not accept new value. */
		getValuesFromLabels: function (labelsData, readOnly) {
			result = []; notFound = [];
			result = this._flatSearch(labelsData, 'label', notFound);
			// Possibility to add the value
			if (notFound.length > 0 && readOnly != undefined && readOnly) {
				for (var i = 0; i < notFound.length; i++) {
					res = {};
					res.value = notFound[i];
					res.label = notFound[i];
					res.html = '';
					result.push(res);
					this._updateSource(this.dataSource, res);
				}
			}
			return result;
		},

		// Search Function with Base Data and Ajax Management
		research: function (searchString, origin, forceAjax) {
			if ((searchString == '' || searchString == undefined) && (forceAjax == undefined || forceAjax == false)) return;
			// Shared Source : Callbacks Target needs to be reset
			if (origin != undefined && origin != '') {
				this._setOrigin(origin);
			}
			// forced Ajax loading
			if (forceAjax == true) {
				if (this.ajax.action != undefined) {
					this._remoteSource('');
				} else {
					//if no ajax available, send back original datasource
					this._sendBack(this.dataSource, "ok");
				}
			} else {
				// searchString too short
				if (searchString.length < this.options.minlength) return;
				// ajax has already been loaded
				if (searchString.toLowerCase().substring(0, this.currentOrigin.searched.length) == this.currentOrigin.searched.toLowerCase) {
					resultData = this._recursiveSearch(searchString, this.dataSource, false);
					if (resultData.length < this.options.minoption && (this.ajax.sourceurl != undefined || this.ajax.action != undefined)) {
						this.currentOrigin.searched = searchString;
						this.currentOrigin.page = 0;
						this._remoteSource(searchString);
					} else {
						this._sendBack(resultData, "ok")
					}
				} else {
					// ajax has not been loaded or new searchString
					this.currentOrigin.searched = searchString;
					this.currentOrigin.page = 0;
					if (searchString.length > this.options.minsearch && (this.ajax.sourceurl != undefined || this.ajax.action != undefined)) {
						this._remoteSource(searchString);
					} else {
						// launch standard Search
						resultData = this._recursiveSearch(searchString, this.dataSource, false);
						// too few standard search results, load ajax data
						if (resultData.length < this.options.minoption && (this.ajax.sourceurl != undefined || this.ajax.action != undefined)) {
							this._remoteSource(searchString);
						} else {
							// send search result to callback
							this._sendBack(resultData, "ok")
						}
					}
				}
			}
		},

		// return the dataSource as a result
		returnSource: function () {
			this._sendBack(this.dataSource, "ok");
		},

		// send the complete source to view
		getSource: function () {
			return this._mapToArray(this.dataSource);
		},

		// add Custom Elements to dataSource
		addToSource: function (data) {
			this._updateSource(this.dataSource, data);
		},

		clone: function (obj) {
			if (obj == null || typeof (obj) != 'object')
				return obj;
			if (obj instanceof Map)
				return new Map(obj);
			var temp = new obj.constructor();
			for (var key in obj)
				temp[key] = this.clone(obj[key]);
			return temp;
		},

		/* Search in the flatSource */
		_flatSearch: function (searched, from, notFound) {
			var resultData = [];
			var me = this;

			$.each(searched, function (i, searchedValue) {
				var found = false;
				if ('value' == from) {
					item = me.flatSource.get(searchedValue.toString());
					if (undefined != item) {
						resultData.push(me.clone(item));
						found = true;
					}
				} else for (const [j, sourceValue] of me.flatSource) {
					if (searchedValue == sourceValue.label) {
						resultData.push(me.clone(sourceValue));
						found = true;
						break;
					}
				}

				if (!found)
					notFound.push(searchedValue);
			});
			return resultData;
		},

		//Recursive Search Function
		_recursiveSearch: function (searched, dataSource, fromVal) {
			var resultData = new Map();
			if (dataSource.size == 0) return resultData;
			for (const [key, item] of dataSource) {
				/* optgroups */
				if (item.group != undefined) {
					//retour = {};
					retour = this._recursiveSearch(searched, item.group, fromVal);
					if (retour.size > 0) {
						resultData.set(key, this.clone(item));
						resultData[key].set('group', retour);
					}
				} else {
					if (fromVal == true) {
						for (var i = 0; i < searched.length; i++) {
							if (searched[i] == item.value) {
								resultData.set(key, this.clone(item));
								this.resCount -= 1;
							};
						}
					} else {
						researched = item.label.toLowerCase();
						if ($.isArray(searched)) {
							for (let i = 0; i < searched.length; i++) {
								if (this.options.mode == 'autocomplete' || this.options.mode == 'select') {
									if (researched.substring(0, searched.length) == searched[i].toLowerCase()) {
										resultData.set(key, this.clone(item));
										this.resCount -= 1;
									}
								} else if (this.options.mode == 'research') {
									if (researched.search(searched.toLowerCase()) != -1) {
										resultData.set(key, this.clone(item));
										this.resCount -= 1;
									}
								}
							}
						} else {
							if (this.options.mode == 'autocomplete' || this.options.mode == 'select') {
								if (researched.substring(0, searched.length) == searched.toLowerCase()) {
									resultData.set(key, this.clone(item));
								}
							} else if (this.options.mode == 'research') {
								if (researched.search(searched.toLowerCase()) != -1) {
									resultData.set(key, this.clone(item));
								}
							}
						}
					}
				}
			}
			return resultData;
		},

		// Updates the dataSources with incoming data
		_updateSource: function (dataSource, newSource, from, me) {
			if (undefined == me)
				me = this;
			$.each(newSource, function (index, sourceItem) {
				if (sourceItem.group != undefined) {
					var label = sourceItem.label.toString();
					var keys = undefined;
					// It is a group entry
					if (me.groupSource.has(label)) {
						/* Group already exists */
						keys = me._updateSource(me.groupSource.get(label).group, sourceItem.group, label, me);
					} else {
						var grp = { 'label': label, 'group': new Map() };
						dataSource.set(sourceItem.label.toString(), grp);
						me.groupSource.set(sourceItem.label.toString(), grp); // store group as flat array, whatever the deep
						keys = me._updateSource(grp.group, sourceItem.group, label, me);
					}
					// remove any pre-loaded value put a root by default
					me.groupSource.get(label).group.forEach(function (v, k, m) {
						dataSource.delete(k);
					});
				} else if (sourceItem.value != undefined) {
					var value = sourceItem.value.toString();
					/* It's a value, add it or update it in the datasource */
					if (undefined !== from || !me.flatSource.has(value)) {
						me.flatSource.set(value, sourceItem);
						dataSource.set(value, sourceItem);
					}
				}
			});
			return dataSource.keys();
		},

		/** Fo async purpose, call the callback function registered on resChange
		 *	With an array [data, status, showmore] where data is
		 *	an array of object {value, label, html, group}
		 * */
		_sendBack: function (resultData, status) {
			if (resultData instanceof Map)
				resultData = this._mapToArray(resultData);
			result = [resultData, status, this.showMore];
			if (this.currentOrigin.resChange != undefined) {
				this.currentOrigin.resChange.fn.apply(this.currentOrigin.target, [result]);
			}
		},

		/** Recursively transform Map to array (dig in group value) */
		/** Change function to array without references */
		_mapToArray: function (m) {
			var me = this;
			var a = [];
			m.forEach(function (obj, k, map) {
				let item = {};
				for (const [key, value] of Object.entries(obj)) {
					if (key != 'group')
						item[key] = value;
					else
						item.group = me._mapToArray(value);
				}
				a.push(item);
			});
			return a;
		},

		// Calculates the length of an ajax result
		_getResultLength: function (items) {
			length = 0;
			for (var key in items) {
				length += (items[key].group) ? Object.keys(items[key].group).length : 1;
			}
			return length;
		},

		// retrieve the ajax data set in parameter and returns an array
		_remoteSource: function (searched, initStatus, fromVal) {
			if (this.currentOrigin.resLoad != undefined)
				this.currentOrigin.resLoad.fn.apply(this.currentOrigin.target);
			var currentPage = this.currentOrigin.page;
			var args = {
				action: this.ajax.action,
				term: searched,
				fromValue: fromVal ? 'yes' : '',
				count: this.currentOrigin.count,
				page: currentPage
			};
			//if (initStatus == 'init') args.term = ''; args.fromValue = undefined;

			if (this.currentOrigin.target.element.data('spec') != undefined)
				args.spec = this.currentOrigin.target.element.data('spec');

			ajax_url = (this.ajax.sourceurl != undefined) ? this.ajax.sourceurl : lws_ajax.url;
			var me = this;
			$.getJSON(ajax_url, args, function (items) {
				resultData = [];
				if (null == items) status = "Autocomplete ressource error.";
				else if (0 === items) status = "Autocomplete action not found.";
				else {
					me._updateSource(me.dataSource, items);
					me.showMore = (me._getResultLength(items) == me.currentOrigin.count);
					// Force the ajax load on init if ajax specified.
					if (initStatus == "force") {
						return;
					}
					// initStatus = init --> On renvoie le label correpondant à la clé stockée en bdd
					if (initStatus == "init") {
						status = "init";
						notFound = [];
						resultData = me._flatSearch(searched, 'value', notFound);
						if (notFound.length > 0) {
							console.log("some items couldn't be found on datasource");
							console.log(notFound);
						}
					} else {
						status = "ok";
						me.currentOrigin.searched = searched;
						if (currentPage > 0)
							resultData = me.dataSource;
						else {
							if ('object' == typeof items) {
								for (prop in items)
									resultData.push(items[prop]);
							}
						}
					}
					me._sendBack(resultData, status);
				}
			}).fail(function (d, textStatus, error) {
				resultData = [];
				status = "Autocomplete loading error, status: " + textStatus + ", error: " + error;
				me._sendBack(resultData, status)
			});
		}

	});

})(jQuery)